;;; smv-hooks.el --- run gdb under XEmacs

;; 
;;
;;  Originally gdb.el, from XEmacs 19.16
;;
;;  Modified by K. L. McMillan to make smv-hooks.el
;;
;;  This code is *not* part of the SMV model checker, but you may
;;  find it useful for running smv under XEmacs.
;;
;;  Original header information from gdb.el follows:
;;  
;; Author: W. Schelter, University of Texas
;;     wfs@rascal.ics.utexas.edu
;; Rewritten by rms.
;; Keywords: c, unix, tools, debugging

;; Some ideas are due to Masanobu.

;; This file is part of XEmacs.

;; XEmacs is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; XEmacs is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with XEmacs; see the file COPYING.  If not, write to the Free
;; Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;; 02111-1307, USA.

;; Usage:
;;
;; (smv-hooks)
;;
;; Modifies shell mode so that errors reported by gdb or
;; smv will be indicated in the source file by an arrow. In fact, this
;; will work with any program that indicates source lines in the same
;; manner as gdb --fullname, and is not specific to smv. It also sets
;; up to use smv-mode (see smv-mode.el) for file names ending in ".smv".
;; Finally, it calls the function "gnuserv-start" if it exists, so that
;; XEmacs will respond to requests from "vw" to open files. If you are
;; security-minded, you might want to remove this call.


;;; Code:

(require 'shell)

(defvar my-last-frame)
(defvar my-delete-prompt-marker)
(defvar my-filter-accumulator)
(defvar my-last-frame-displayed-p)
(defvar current-my-buffer nil)

(setq shell-mode-hook '(smv-shell-mode-hook))
(setq auto-mode-alist (cons '("\\.smv$" . smv-mode) auto-mode-alist))
(autoload 'smv-mode "smv-mode")

(defun smv-shell-mode-hook ()
  (mapcar 'make-local-variable
 	  '(my-last-frame-displayed-p  my-last-frame
				       my-delete-prompt-marker    my-filter-accumulator))
  (setq
   my-last-frame nil
   my-delete-prompt-marker nil
   my-filter-accumulator nil
   )
  (set-process-filter (get-buffer-process (current-buffer)) 'my-filter)
  (set-process-sentinel (get-buffer-process (current-buffer)) 'my-sentinel)
  (my-set-buffer))

(defun my-set-buffer ()
  (cond ((eq major-mode 'shell-mode)
	 (setq current-my-buffer (current-buffer)))))


;; This function is responsible for inserting output from MY
;; into the buffer.
;; Aside from inserting the text, it notices and deletes
;; each filename-and-line-number;
;; that MY prints to identify the selected frame.
;; It records the filename and line number, and maybe displays that file.
(defun my-filter (proc string)
  (let ((inhibit-quit t))
    (save-current-buffer
     (set-buffer (process-buffer proc))
     (if my-filter-accumulator
	 (my-filter-accumulate-marker
	  proc (concat my-filter-accumulator string))
       (my-filter-scan-input proc string)))))


(defun my-filter-accumulate-marker (proc string)
  (setq my-filter-accumulator nil)
  (if (> (length string) 1)
      (if (= (aref string 1) ?\032)
 	  (let ((end (string-match "\n" string)))
 	    (if end
 		(progn
 		  (let* ((first-colon (string-match ":" string 2))
 			 (second-colon
 			  (string-match ":" string (1+ first-colon))))
 		    (setq my-last-frame
 			  (cons (substring string 2 first-colon)
 				(string-to-int
 				 (substring string (1+ first-colon)
 					    second-colon)))))
 		  (setq my-last-frame-displayed-p nil)
 		  (my-filter-scan-input proc
					(substring string (1+ end))))
 	      (setq my-filter-accumulator string)))
 	(my-filter-insert proc "\032")
 	(my-filter-scan-input proc (substring string 1)))
    (setq my-filter-accumulator string)))

(defun my-filter-scan-input (proc string)
  (if (equal string "")
      (progn
 	(setq my-filter-accumulator nil)
 	(my-filter-insert proc string))
    (let ((start (string-match "\032" string)))
      (if start
	  (progn (my-filter-insert proc (substring string 0 start))
		 (my-filter-accumulate-marker proc
					      (substring string start)))
	(my-filter-insert proc string)))))

(defun my-filter-insert (proc string)
  (comint-output-filter proc string)
  (let (
;	(moving (= (point) (process-mark proc)))
 	(output-after-point (< (point) (process-mark proc)))
; 	(old-buffer (current-buffer))
; 	start
	)
;    (set-buffer (process-buffer proc))
;    (unwind-protect
; 	(save-excursion
; 	  ;; Insert the text, moving the process-marker.
; 	  (goto-char (process-mark proc))
; 	  (setq start (point))
; 	  (insert-before-markers string)
; 	  (set-marker (process-mark proc) (point))
; 	  ;;(my-maybe-delete-prompt)
; 	  ;; Check for a filename-and-line number.
 	  (my-display-frame
 	   ;; Don't display the specified file
 	   ;; unless (1) point is at or after the position where output appears
 	   ;; and (2) this buffer is on the screen.
 	   (or output-after-point
 	       (not (get-buffer-window (current-buffer))))
 	   ;; Display a file only when a new filename-and-line-number appears.
;;	   nil
 	   t))
;      (set-buffer old-buffer))
;    (if moving (goto-char (process-mark proc))))
)

(defun my-sentinel (proc msg)
  (cond ((null (buffer-name (process-buffer proc)))
 	 ;; buffer killed
 	 ;; Stop displaying an arrow in a source file.
 	 (setq overlay-arrow-position nil)
 	 (set-process-buffer proc nil))
 	((memq (process-status proc) '(signal exit))
 	 ;; Stop displaying an arrow in a source file.
 	 (setq overlay-arrow-position nil)
 	 ;; Fix the mode line.
 	 (setq mode-line-process
 	       (concat ": "
 		       (symbol-name (process-status proc))))
 	 (let* ((obuf (current-buffer)))
 	   ;; save-excursion isn't the right thing if
 	   ;;  process-buffer is current-buffer
 	   (unwind-protect
 	       (progn
 		 ;; Write something in *compilation* and hack its mode line,
 		 (set-buffer (process-buffer proc))
 		 ;; Force mode line redisplay soon
 		 (set-buffer-modified-p (buffer-modified-p))
 		 (if (eobp)
 		     (insert ?\n mode-name " " msg)
 		   (save-excursion
 		     (goto-char (point-max))
 		     (insert ?\n mode-name " " msg)))
 		 ;; If buffer and mode line will show that the process
 		 ;; is dead, we can delete it now.  Otherwise it
 		 ;; will stay around until M-x list-processes.
 		 (delete-process proc))
 	     ;; Restore old buffer, but don't restore old point
 	     ;; if obuf is the my buffer.
 	     (set-buffer obuf))))))


(defun my-refresh (&optional arg)
  "Fix up a possibly garbled display, and redraw the arrow."
  (interactive "P")
  (recenter arg)
  (my-display-frame))

(defun my-display-frame (&optional nodisplay noauto)
  "Find, obey and delete the last filename-and-line marker from MY.
 The marker looks like \\032\\032FILENAME:LINE:CHARPOS\\n.
 Obeying it means displaying in another window the specified file and line."
  (interactive)
  (my-set-buffer)
  (and my-last-frame (not nodisplay)
       (or (not my-last-frame-displayed-p) (not noauto))
       (progn (my-display-line (car my-last-frame) (cdr my-last-frame))
 	      (setq my-last-frame-displayed-p t))))

;; Make sure the file named TRUE-FILE is in a buffer that appears on the screen
;; and that its line LINE is visible.
;; Put the overlay-arrow on the line LINE in that buffer.

(defun my-display-line (true-file line)
  (let* ((pre-display-buffer-function nil) ; screw it, put it all in one screen
 	 (pop-up-windows t)
 	 (buffer (find-file-noselect true-file))
 	 (window (display-buffer buffer t))
 	 (pos))
    (save-excursion
      (set-buffer buffer)
      (save-restriction
 	(widen)
 	(goto-line line)
 	(setq pos (point))
 	(setq overlay-arrow-string "=>")
 	(or overlay-arrow-position
 	    (setq overlay-arrow-position (make-marker)))
 	(set-marker overlay-arrow-position (point) (current-buffer)))
      (cond ((or (< pos (point-min)) (> pos (point-max)))
 	     (widen)
 	     (goto-char pos))))
    (set-window-point window overlay-arrow-position)))

(if (fboundp 'gnuserv-start)
    (gnuserv-start))
